#include "ocf/ocf.h"
#include "../ocf_cache_priv.h"
#include "../ocf_queue_priv.h"
#include "../utils/utils_cache_line.h"
#include "../metadata/metadata.h"

#include "../xcache.h"
#include "xcache_engine_common.h"
#include "../xcache_lru.h"
#include "../xcache_queue.h"
#include "../evicting/evicting_ops.h"

uint64_t cache_line_to_addr(ocf_cache_t cache, ocf_cache_line_t line, uint64_t line_offset)
{
	uint64_t addr;

	addr = ocf_metadata_map_lg2phy(cache, line);
	addr = xcache_line_to_addr(cache, addr);
	addr += cache->device->metadata_offset;
	addr += line_offset;
	return addr;
}

ocf_cache_line_t addr_to_cache_line(ocf_cache_t cache, uint64_t addr)
{
	ocf_cache_line_t line;

	addr -= cache->device->metadata_offset;
	line = xcache_addr_to_line(cache, addr);
	line = ocf_metadata_map_phy2lg(cache, line);
	return line;
}

ocf_cache_line_t xcache_engine_lookup_line(struct ocf_cache *cache,
		ocf_core_id_t core_id, uint64_t core_line)
{
	ocf_cache_line_t line;
	ocf_cache_line_t hash;

	hash = ocf_metadata_hash_func(cache, core_line, core_id);
	line = ocf_metadata_get_hash(cache, hash);

	while (line != cache->device->collision_table_entries) {
		ocf_core_id_t curr_core_id;
		uint64_t curr_core_line;

		ocf_metadata_get_core_info(cache, line, &curr_core_id,
				&curr_core_line);

		if (core_id == curr_core_id && curr_core_line == core_line) {
			return line;
		}

		line = ocf_metadata_get_collision_next(cache, line);
	}

	return INVALID_LINE;
}

void xcache_map_cache_line(struct xcache_io_context *ctx,
		ocf_cache_line_t cache_line)
{
	ocf_cache_t cache = xcache_ctx_cache(ctx);
	ocf_core_id_t core_id = ocf_core_get_id(xcache_ctx_core(ctx));
	unsigned int hash_index = ctx->hash;
	uint64_t core_line = ctx->core_line;

	/* Add the block to the corresponding collision list */
	ocf_metadata_start_collision_shared_access(cache, cache_line);
	ocf_metadata_add_to_collision(cache, core_id, core_line, hash_index,
			cache_line);
	ocf_metadata_end_collision_shared_access(cache, cache_line);
}

static void xcache_engine_prepare_wait(void *io)
{
	struct xcache_backdev_io *base_io = (struct xcache_backdev_io *)io;
	struct xcache_io *ocf_io = base_io->xcache_io;

	// avoid xcache_io finish when waiting for lock
	xcache_io_get(ocf_io);
}

static void xcache_engine_prepare_wake(void *io)
{
	struct xcache_backdev_io *base_io = (struct xcache_backdev_io *)io;

	xcache_queue_push_backdev_io_front(base_io, false);
}

static void xcache_get_core_range(struct xcache_io_context *ctx,
		uint64_t start_core_line, uint64_t end_core_line,
		uint64_t *addr, uint64_t *size)
{
	struct xcache_io *io = ctx->io;
	ocf_cache_t cache = xcache_ctx_cache(ctx);
	uint64_t start_addr, end_addr, end;

	start_addr = xcache_io_start_addr(io);
	end_addr = start_addr + xcache_io_size(io);

	*addr = xcache_line_to_addr(cache, start_core_line);
	end = xcache_line_to_addr(cache, end_core_line);

	if (*addr < start_addr) {
		*addr = start_addr;
	}
	if (end > end_addr) {
		end = end_addr;
	}

	*size = end - *addr;
}

static inline int xcache_engine_lock_wr(struct xcache_io_context *ctx, struct ocf_alock *alock,
		uint64_t core_line, ocf_cache_line_t line,
		backdev_io_res_fn io_res)
{
	ocf_queue_t q = xcache_ctx_queue(ctx);
	struct xcache_backdev_io *base_io = xcache_queue_alloc_backdev_io(q);
	int ret;

	if (base_io == NULL) {
		ocf_cache_log(xcache_ctx_cache(ctx), log_err, "alloc base io failed\n");
		return -ENOMEM;
	}

	base_io->xcache_io = ctx->io;
	base_io->line = line;
	base_io->io_res = io_res;
	base_io->data = ctx->io->data;
	xcache_get_core_range(ctx, core_line, core_line + 1, &base_io->addr, &base_io->size);

	ret = xcache_lock_wr(alock, line, xcache_engine_prepare_wait,
			xcache_engine_prepare_wake, base_io);
	if (ret != OCF_LOCK_NOT_ACQUIRED) {
		xcache_queue_free_backdev_io(q, base_io);
	}

	return ret;
}

static inline void xcache_hash_lock_rd(ocf_cache_t cache, ocf_cache_line_t hash, uint8_t lock_idx)
{
	ocf_metadata_start_shared_access(&cache->metadata.lock,
			lock_idx);
	ocf_hb_id_naked_lock(&cache->metadata.lock, hash,
			OCF_METADATA_RD);
}

static inline void xcache_hash_unlock_rd(ocf_cache_t cache, ocf_cache_line_t hash, uint8_t lock_idx)
{
	ocf_hb_id_naked_unlock(&cache->metadata.lock, hash,
			OCF_METADATA_RD);
	ocf_metadata_end_shared_access(&cache->metadata.lock,
			lock_idx);
}

static inline void xcache_hash_lock_upgrade(ocf_cache_t cache, ocf_cache_line_t hash)
{
	ocf_hb_id_naked_unlock(&cache->metadata.lock, hash,
			OCF_METADATA_RD);
	ocf_hb_id_naked_lock(&cache->metadata.lock, hash,
			OCF_METADATA_WR);
}

static inline void xcache_hash_unlock_wr(ocf_cache_t cache, ocf_cache_line_t hash, uint8_t lock_idx)
{
	ocf_hb_id_naked_unlock(&cache->metadata.lock, hash,
			OCF_METADATA_WR);
	ocf_metadata_end_shared_access(&cache->metadata.lock,
			lock_idx);
}

int xcache_engine_get_line(struct xcache_io_context *ctx,
		uint64_t core_line, ocf_cache_line_t *line,
		backdev_io_res_fn io_res)
{
	ocf_cache_t cache = xcache_ctx_cache(ctx);
	ocf_core_t core = xcache_ctx_core(ctx);
	ocf_queue_t q = xcache_ctx_queue(ctx);
	ocf_core_id_t core_id = ocf_core_get_id(core);
	unsigned lock_idx = ocf_metadata_concurrency_next_idx(q);
	ocf_cache_line_t hash = ocf_metadata_hash_func(cache, core_line, core_id);
	struct ocf_alock *alock = ocf_cache_line_concurrency(cache);

	int lock;

	xcache_hash_lock_rd(cache, hash, lock_idx);
	*line = xcache_engine_lookup_line(cache, core_id, core_line);
	if (*line != INVALID_LINE) {
		lock = xcache_engine_lock_wr(ctx, alock, core_line, *line, io_res);
		xcache_hash_unlock_rd(cache, hash, lock_idx);
		goto found;
	}

	xcache_hash_lock_upgrade(cache, hash);
	*line = xcache_engine_lookup_line(cache, core_id, core_line);
	if (unlikely(*line) != INVALID_LINE) {
		lock = xcache_engine_lock_wr(ctx, alock, core_line, *line, io_res);
		xcache_hash_unlock_wr(cache, hash, lock_idx);
		goto found;
	}

	ctx->core_line = core_line;
	ctx->hash = hash;
	*line = xcache_get_cline(ctx);
	lock = OCF_LOCK_ACQUIRED;
	xcache_hash_unlock_wr(cache, hash, lock_idx);
	if (*line == INVALID_LINE) {
		xcache_clean(ctx, 128);
		return lock;
	}

found:
	evicting_line_accessed(xcache_get_ctx(cache), *line);
	return lock;
}

void xcache_init_io_ctx(struct xcache_io_context *ctx, struct xcache_io *io)
{
	uint64_t addr = xcache_io_start_addr(io);
	uint64_t size = xcache_io_size(io);

	ctx->io = io;
	ctx->part_id = PARTITION_DEFAULT;
	ctx->offset = 0;
	ctx->start_miss_line = INVALID_CORE_LINE;
	ctx->hit_no = 0;
	ctx->cache_bytes = 0;
	ctx->core_bytes = 0;
	xcache_io_get_line_range(xcache_io_cache(io), addr, size,
			&ctx->core_line_first, &ctx->core_line_last);
}

void xcache_update_stat(struct xcache_io_context *ctx)
{
	struct xcache_io *io = ctx->io;
	uint64_t len = xcache_io_size(io);

	ocf_core_stats_request_update(xcache_ctx_core(ctx), PARTITION_DEFAULT,
			io->rw, ctx->hit_no, ctx->core_line_last - ctx->core_line_first + 1);
	ocf_core_stats_vol_block_update(xcache_ctx_core(ctx), PARTITION_DEFAULT,
			io->rw, len);
	ocf_core_stats_core_block_update(xcache_ctx_core(ctx), PARTITION_DEFAULT,
			io->rw, ctx->core_bytes);
	ocf_core_stats_cache_block_update(xcache_ctx_core(ctx), PARTITION_DEFAULT,
			io->rw, ctx->cache_bytes);
}

static void xcache_queue_miss_line(struct xcache_io_context *ctx, uint64_t core_line)
{
	if (ctx->start_miss_line == INVALID_CORE_LINE) {
		ctx->start_miss_line = core_line;
	}
}

static int xcache_submit_miss_line(struct xcache_io_context *ctx, uint64_t end_miss_line, xcache_line_miss_fn submit_func)
{
	struct xcache_io *io = ctx->io;

	struct xcache_backdev_io *base_io;
	uint64_t start_addr, end_addr;
	uint64_t addr, end;

	if (submit_func == NULL) {
		ctx->start_miss_line = INVALID_CORE_LINE;
		return 0;
	}

	if (ctx->start_miss_line == INVALID_CORE_LINE) {
		return 0;
	}

	base_io = xcache_queue_alloc_backdev_io(xcache_ctx_queue(ctx));
	if (base_io == NULL) {
		ocf_cache_log(xcache_ctx_cache(ctx), log_err, "alloc base io failed\n");
		io->error = -ENOMEM;
		return -ENOMEM;
	}
	base_io->xcache_io = io;
	base_io->data = io->data;

	start_addr = xcache_io_start_addr(io);
	end_addr = start_addr + xcache_io_size(io);
	addr = xcache_line_to_addr(xcache_ctx_cache(ctx), ctx->start_miss_line);
	end = xcache_line_to_addr(xcache_ctx_cache(ctx), end_miss_line);
	if (addr < start_addr) {
		addr = start_addr;
	}
	if (end > end_addr) {
		end = end_addr;
	}

	submit_func(base_io, addr, end - addr, addr - start_addr);

	ctx->start_miss_line = INVALID_CORE_LINE;
	return 0;
}

int xcache_foreach_line(struct xcache_io_context *ctx, xcache_line_handle_func func, void *priv)
{
	int ret = 0;
	struct xcache_io *io = ctx->io;
	uint64_t start_addr = xcache_io_start_addr(io);
	uint64_t len = xcache_io_size(io);

	uint64_t addr, line_end, line_size, size;
	uint8_t start_sector, end_sector, line_sectors;
	uint64_t core_line_first, core_line_last, core_line_cur;

	core_line_first = ctx->core_line_first;
	core_line_last = ctx->core_line_last;
	addr = start_addr;
	line_size = ocf_line_size(xcache_ctx_cache(ctx));
	line_sectors = BYTES_TO_SECTORS(line_size);
	line_end = xcache_line_to_addr(xcache_ctx_cache(ctx), core_line_first + 1);
	start_sector = xcache_sector_offset(xcache_ctx_cache(ctx), BYTES_TO_SECTORS(addr));

	for (core_line_cur = core_line_first;
			core_line_cur <= core_line_last;
			core_line_cur++) {
		if (core_line_cur == core_line_last) {
			end_sector = xcache_sector_offset(xcache_ctx_cache(ctx), BYTES_TO_SECTORS(start_addr + len - 1));
			size = start_addr + len - addr;
		} else {
			end_sector = line_sectors - 1;
			size = line_end - addr;
		}

		ret = func(priv, core_line_cur,
				addr, size, start_sector, end_sector);
		if (ret) {
			io->error = ret;
			break;
		}
		ctx->offset += size;
		addr += size;
		line_end += line_size;
		start_sector = 0;
	}

	return 0;
}

static int xcache_wr_lb_common_end(struct xcache_backdev_io *backdev_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = backdev_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_cache_line_t line = backdev_io->line;
	uint8_t start_secotr, last_sector;

	xcache_get_sectors_range(cache, cb_arg->addr, cb_arg->size, &start_secotr, &last_sector);
	ocf_metadata_start_collision_shared_access(cache, line);
	metadata_clear_valid_sec(cache, line, start_secotr, last_sector);
	ocf_metadata_end_collision_shared_access(cache, line);

	xcache_unlock_wr(ocf_cache_line_concurrency(cache), line);
	xcache_queue_free_backdev_io(xcache_io_queue(backdev_io->xcache_io), backdev_io);
	return 0;
}

void xcache_wr_lb_common(struct xcache_backdev_io *backdev_io, ocf_cache_line_t line,
		uint64_t addr, uint64_t size)
{
	struct xcache_io *io = backdev_io->xcache_io;
	uint64_t start_addr = xcache_io_start_addr(io);
	uint64_t buf_offset = addr - start_addr;

	backdev_io->line = line;
	backdev_io->end = xcache_wr_lb_common_end;
	xcache_backdev_submit_io(backdev_io, false, addr, size, buf_offset, OCF_WRITE);
}

static inline bool xcache_engine_need_lb(struct xcache_io_handler *handler, struct xcache_line_range *line_range)
{
	struct xcache_io_context *ctx = handler->ctx;
	struct xcache_io *io = ctx->io;
	ocf_cache_t cache = xcache_io_cache(io);
	xcache_context_t *xcache_ctx = xcache_get_ctx(cache);
	struct backdev_io *backdev_io = NULL;

	if (handler->lb_fn == NULL) {
		return false;
	}

	if (handler->need_lb_fn != NULL &&
			!handler->need_lb_fn(xcache_io_cache(io), line_range)) {
		return false;
	}

	return xcache_qos_need_lb(&xcache_ctx->qos, xcache_io_dir(io), line_range->size);
}

static int xcache_handle_line(void *priv,
		uint64_t core_line, uint64_t addr, uint64_t size,
		uint8_t start_sector, uint8_t last_sector)
{
	struct xcache_io_handler *handler = (struct xcache_io_handler *)priv;
	struct xcache_io_context *ctx = handler->ctx;
	struct xcache_io *io = ctx->io;
	ocf_cache_t cache = xcache_ctx_cache(ctx);
	struct xcache_backdev_io *backdev_io;
	struct xcache_line_range line_range = {
		.addr = addr,
		.size = size,
		.start_sector = start_sector,
		.last_sector = last_sector,
	};

	ocf_cache_line_t line;
	int lock;

	lock = xcache_engine_get_line(ctx, core_line, &line, handler->res_fn);
	if (lock < 0) {
		ocf_cache_log(cache, log_err, "try to wait for lock failed\n");
		return -ENOMEM;
	}

	if (line == INVALID_LINE || (lock == OCF_LOCK_ACQUIRED &&
		handler->valid_fn != NULL && handler->valid_fn(cache, line, start_sector, last_sector))) {
		xcache_queue_miss_line(ctx, core_line);
		ctx->core_bytes += size;
		return 0;
	}

	xcache_submit_miss_line(ctx, core_line, handler->miss_fn);

	if (lock == OCF_LOCK_NOT_ACQUIRED) {
		ctx->hit_no++;
		ctx->cache_bytes += size;
		return 0;
	}

	backdev_io = xcache_alloc_backdev_io(io);
	if (backdev_io == NULL) {
		ocf_cache_log(cache, log_err, "alloc base io failed\n");
		return -ENOMEM;
	}

	line_range.cache_line = line;
	if (xcache_engine_need_lb(handler, &line_range))
	{
		ctx->core_bytes += size;
		handler->lb_fn(backdev_io, line, addr, size);
		return 0;
	}

	ctx->hit_no++;
	ctx->cache_bytes += size;
	handler->hit_fn(backdev_io, line, SECTORS_TO_BYTES(start_sector), size, ctx->offset);
	return 0;
}

int xcache_handle_io(struct xcache_io_handler *handler)
{
	struct xcache_io_context *ctx = handler->ctx;
	int ret;

	ret = xcache_foreach_line(ctx, xcache_handle_line, (void *)handler);
	if (ret != 0) {
		return ret;
	}
	return xcache_submit_miss_line(ctx, ctx->core_line_last + 1, handler->miss_fn);
}
