#ifndef XCACHE_ENGINE_COMMON_H_
#define XCACHE_ENGINE_COMMON_H_

#include "../ocf_cache_priv.h"
#include "../utils/utils_cache_line.h"
#include "../ocf_def_priv.h"

#include "../xcache.h"
#include "../xcache_queue.h"

#define INVALID_CORE_LINE ((uint64_t)-1)
#define INVALID_LINE ((ocf_cache_line_t)-1)

struct xcache_io_context {
	struct xcache_io *io;
	ocf_queue_t queue;
	ocf_part_id_t part_id;
	uint64_t offset;
	uint64_t core_line_first;
	uint64_t core_line_last;
	uint64_t core_line;
	ocf_cache_line_t hash;

	uint64_t start_miss_line;

	uint64_t hit_no;
	uint64_t cache_bytes;
	uint64_t core_bytes;
};

static inline uint8_t xcache_addr_offset(ocf_cache_t cache, uint64_t addr)
{
	return addr & (ocf_line_size(cache) - 1);
}

static inline uint8_t xcache_sector_offset(ocf_cache_t cache, uint64_t sector)
{
	return sector & (ocf_line_sectors(cache) - 1);
}

static inline void xcache_get_sectors_range(ocf_cache_t cache, uint64_t addr, uint64_t size, uint8_t *start_sector, uint8_t *last_sector)
{
	uint64_t offset = xcache_addr_offset(cache, addr);
	*start_sector = BYTES_TO_SECTORS(offset);
	*last_sector = BYTES_TO_SECTORS(offset + size - 1);
}

static inline void xcache_io_get_line_range(ocf_cache_t cache, uint64_t addr, uint64_t size,
		uint64_t *line_first, uint64_t *line_last)
{
	*line_first = xcache_addr_to_line(cache, addr);
	*line_last = xcache_addr_to_line(cache, addr + size - 1);
}

ocf_cache_line_t xcache_engine_lookup_line(struct ocf_cache *cache,
		ocf_core_id_t core_id, uint64_t core_line);
int xcache_engine_get_line(struct xcache_io_context *ctx,
		uint64_t core_line, ocf_cache_line_t *line,
		backdev_io_res_fn io_res);

void xcache_map_cache_line(struct xcache_io_context *ctx,
		ocf_cache_line_t cache_line);

uint64_t cache_line_to_addr(ocf_cache_t cache, ocf_cache_line_t line, uint64_t line_offset);
ocf_cache_line_t addr_to_cache_line(ocf_cache_t cache, uint64_t addr);

struct xcache_line_range {
	ocf_cache_line_t cache_line;
	uint64_t core_line;
	uint64_t addr;
	uint64_t size;
	uint64_t start_sector;
	uint64_t last_sector;
};

typedef int (*xcache_line_handle_func)(void *priv,
		uint64_t core_line, uint64_t addr, uint64_t size,
		uint8_t start_sector, uint8_t last_sector);

void xcache_init_io_ctx(struct xcache_io_context *ctx, struct xcache_io *io);
int xcache_foreach_line(struct xcache_io_context *ctx, xcache_line_handle_func func, void *priv);
void xcache_update_stat(struct xcache_io_context *ctx);

static inline uint64_t xcache_io_start_addr(struct xcache_io *io)
{
	return io->start_addr;
}

static inline uint64_t xcache_io_size(struct xcache_io *io)
{
	return io->size;
}

static inline uint8_t xcache_io_flags(struct xcache_io *io)
{
	return io->flags;
}

static inline ocf_core_t xcache_io_core(struct xcache_io *io)
{
	return io->core;
}

static inline ocf_cache_t xcache_io_cache(struct xcache_io *io)
{
	return io->cache;
}

static inline ocf_queue_t xcache_io_queue(struct xcache_io *io)
{
	return io->io_queue;
}

static inline int xcache_io_dir(struct xcache_io *io)
{
	return io->rw;
}

static inline ocf_core_t xcache_ctx_core(struct xcache_io_context *ctx)
{
	return xcache_io_core(ctx->io);
}

static inline ocf_cache_t xcache_ctx_cache(struct xcache_io_context *ctx)
{
	return xcache_io_cache(ctx->io);
}

static inline ocf_queue_t xcache_ctx_queue(struct xcache_io_context *ctx)
{
	return xcache_io_queue(ctx->io);
}

static inline struct xcache_backdev_io *xcache_alloc_backdev_io(struct xcache_io *io)
{
	struct xcache_backdev_io *backdev_io = xcache_queue_alloc_backdev_io(xcache_io_queue(io));

	if (backdev_io == NULL) {
		return NULL;
	}

	backdev_io->xcache_io = io;
	backdev_io->data = io->data;
	return backdev_io;
}

typedef int (*xcache_line_valid_fn)(ocf_cache_t cache, ocf_cache_line_t line,
		uint8_t start_sector, uint8_t last_sector);
typedef int (*xcache_line_hit_fn)(struct xcache_backdev_io *base_io, ocf_cache_line_t line,
		uint64_t offset, uint64_t size, uint64_t buf_offset);
typedef int (*xcache_line_miss_fn)(struct xcache_backdev_io *base_io, uint64_t addr,
		uint64_t size, uint64_t buf_offset);
typedef int (*xcache_line_need_lb_fn)(ocf_cache_t cache, struct xcache_line_range *line_range);
typedef void (*xcache_line_lb_fn)(struct xcache_backdev_io *base_io, ocf_cache_line_t line,
		uint64_t addr, uint64_t size);

struct xcache_io_handler {
	struct xcache_io_context *ctx;
	xcache_line_valid_fn valid_fn;
	xcache_line_hit_fn hit_fn;
	xcache_line_miss_fn miss_fn;
	backdev_io_res_fn res_fn;
	xcache_line_need_lb_fn need_lb_fn;
	xcache_line_lb_fn lb_fn;
};
int xcache_handle_io(struct xcache_io_handler *handler);

void xcache_wr_lb_common(struct xcache_backdev_io *backdev_io, ocf_cache_line_t line,
		uint64_t addr, uint64_t size);

static inline void xcache_io_get(struct xcache_io *io)
{
	env_atomic_inc_return(&io->remaining);
}

static inline void xcache_io_end(struct xcache_io *io, int error)
{
	if (io->end)
		io->end(io, error);

}

static inline void xcache_io_put(struct xcache_io *io)
{
	if (env_atomic_dec_return(&io->remaining))
		return;

	xcache_io_end(io, io->error);
}

static inline xcache_context_t *backdev_io_to_xcache_ctx(struct xcache_backdev_io *io_base)
{
	struct xcache_io *io = io_base->xcache_io;
	ocf_queue_t q = io->io_queue;
	ocf_cache_t cache = q->cache;
	return xcache_get_ctx(cache);
}

static inline void xcache_backdev_submit_io(struct xcache_backdev_io *io_base, bool cached, uint64_t addr, uint64_t size, uint64_t buf_offset, uint8_t dir)
{
	struct xcache_io *io = io_base->xcache_io;
	xcache_context_t *xcache_ctx = backdev_io_to_xcache_ctx(io_base);

	io_base->dir = dir;
	if (cached) {
		io_base->dev = CACHE_DEV;
		xcache_qos_load_add(&xcache_ctx->qos, CACHE_DEV, dir, size);
	} else {
		io_base->dev = CORE_DEV;
		xcache_qos_load_add(&xcache_ctx->qos, CORE_DEV, dir, size);
	}

	xcache_io_get(io);
	spdk_backdev_submit_io(io_base, cached, addr, size, buf_offset, dir);
}
#endif /* XCACHE_ENGINE_COMMON_H_ */
