#include "ocf/ocf.h"
#include "../ocf_cache_priv.h"
#include "../ocf_queue_priv.h"
#include "../xcache_queue.h"

#include "xcache_engine_common.h"
#include "xcache_engine_flush.h"

env_atomic g_need_flush = 0;

static inline int xcache_do_flush(struct xcache_io *io);

static inline int xcache_push_flush_io(struct xcache_io *io)
{
	ocf_queue_t q = xcache_io_queue(io);
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);

	INIT_LIST_HEAD(&io->queue_list);
	list_add_tail(&io->queue_list, &queue_ctx->flush_io_list);
	env_atomic_inc(&queue_ctx->flush_io_no);
	return env_atomic_read(&queue_ctx->flush_io_no);
}

static inline struct xcache_io *xcache_pop_flush_io(ocf_queue_t q)
{
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);
	struct xcache_io *flush_io = list_first_entry(&queue_ctx->flush_io_list,
			struct xcache_io, queue_list);

	env_atomic_dec(&queue_ctx->flush_io_no);
	list_del(&flush_io->queue_list);

	return flush_io;
}

static void xcache_flush_finish(struct xcache_io *io)
{
	ocf_queue_t q = io->io_queue;
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);
	struct xcache_io *finish_io;
	struct xcache_io *flush_io;

	for (finish_io = xcache_pop_flush_io(q);
			finish_io != io;
			finish_io = xcache_pop_flush_io(q))
	{
		xcache_io_put(finish_io);
	}
	xcache_io_put(io);

	if (env_atomic_read(&queue_ctx->flush_io_no) != 0) {
		flush_io = list_entry(queue_ctx->flush_io_list.prev, struct xcache_io, queue_list);
		xcache_do_flush(flush_io);
	}
}

static int xcache_flush_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);

	if (cb_arg->error != 0) {
		ocf_cache_log(cache, log_err, "flush with error %d\n", cb_arg->error);
		io->error = cb_arg->error;
	}
	xcache_queue_free_backdev_io(xcache_io_queue(io), base_io);

	if (env_atomic_read(&io->remaining) == 2) {
		xcache_flush_finish(io);
	}
	return 0;
}

static int xcache_flush_core(struct xcache_io *io)
{
	uint64_t addr = xcache_io_start_addr(io);
	uint64_t size = xcache_io_size(io);
	ocf_cache_t cache = xcache_io_cache(io);
	struct xcache_backdev_io *base_io = NULL;

	base_io = xcache_queue_alloc_backdev_io(xcache_io_queue(io));
	if (base_io == NULL) {
		ocf_cache_log(cache, log_err, "alloc base io failed\n");
		return -ENOMEM;
	}

	base_io->xcache_io = io;
	base_io->end = xcache_flush_cb;
	xcache_backdev_submit_io(base_io, false, addr, size, 0, OCF_FLUSH);
	return 0;
}

static int xcache_flush_cache(struct xcache_io *io)
{
	ocf_cache_t cache = xcache_io_cache(io);
	struct xcache_backdev_io *base_io = NULL;

	base_io = xcache_queue_alloc_backdev_io(xcache_io_queue(io));
	if (base_io == NULL) {
		ocf_cache_log(cache, log_err, "alloc base io failed\n");
		return -ENOMEM;
	}

	base_io->xcache_io = io;
	base_io->end = xcache_flush_cb;
	xcache_backdev_submit_io(base_io, true, 0, 0, 0, OCF_FLUSH);
	return 0;
}

static inline int xcache_do_flush(struct xcache_io *io)
{
	if (xcache_flush_core(io) != 0) {
		io->error = -ENOMEM;
		return -ENOMEM;
	}

	if (xcache_flush_cache(io) != 0) {
		io->error = -ENOMEM;
		return -ENOMEM;
	}

	return 0;
}

int xcache_flush(struct xcache_io *io)
{
	io->error = 0;

	if (!need_flush()) {
		return 0;
	}
	clear_flush();

	xcache_io_get(io);
	if (xcache_push_flush_io(io) != 1) {
		return 0;
	}

	return xcache_do_flush(io);
}
