#include "ocf/ocf.h"
#include "../ocf_cache_priv.h"
#include "../concurrency/ocf_concurrency.h"
#include "../utils/utils_cache_line.h"
#include "../metadata/metadata.h"
#include "../ocf_def_priv.h"

#include "../xcache.h"
#include "xcache_engine_rd.h"
#include "xcache_engine_common.h"
#include "../xcache_queue.h"

static inline void xcache_read_bf_done(ocf_cache_t cache, struct xcache_backdev_io *base_io)
{
	ocf_queue_t q = xcache_io_queue(base_io->xcache_io);

	xcache_unlock_wr(ocf_cache_line_concurrency(cache), base_io->line);
	xcache_queue_free_line_data(q, cache, base_io->data);
	base_io->data = NULL;
	xcache_queue_free_backdev_io(q, base_io);
}

static int xcache_read_bf_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *bf_io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(bf_io);
	ocf_core_t core = xcache_io_core(bf_io);

	if (cb_arg->error) {
		ocf_cache_log(cache, log_err, "read bf failed\n");
		ocf_core_stats_cache_error_update(core, OCF_WRITE);
	}
	xcache_read_bf_done(cache, base_io);
	return 0;
}

static int xcache_read_bf_update_metadata(struct xcache_backdev_io *base_io, uint64_t addr, uint64_t size)
{
	ocf_core_t core = xcache_io_core(base_io->xcache_io);
	ocf_cache_t cache = ocf_core_get_cache(core);
	ocf_cache_line_t line = base_io->line;
	ocf_part_id_t part_id = PARTITION_DEFAULT;
	uint8_t start_sector = xcache_sector_offset(cache, BYTES_TO_SECTORS(addr));
	uint8_t end_sector = xcache_sector_offset(cache, BYTES_TO_SECTORS(addr + size - 1));

	// set_cache_line_valid
	ocf_metadata_start_collision_shared_access(cache, line);
	if (unlikely(metadata_test_valid_sec(cache, line, start_sector, end_sector))) {
		ocf_metadata_end_collision_shared_access(cache, line);
		return -1;
	}

	if (metadata_set_valid_sec_changed(cache, line, start_sector, end_sector)) {
		/*
		 * Update the number of cached data for that core object
		 */
		env_atomic_inc(&core->runtime_meta->cached_clines);
		env_atomic_inc(&core->runtime_meta->
				part_counters[part_id].cached_clines);
	}
	ocf_metadata_end_collision_shared_access(cache, line);
	return 0;
}

static ctx_data_t *xcache_get_bf_data(struct xcache_io *io, uint64_t addr, uint64_t size)
{
	ocf_cache_t cache = xcache_io_cache(io);
	uint64_t start_byte = xcache_io_start_addr(io);
	uint64_t from = addr - start_byte;
	ctx_data_t *dst;

	dst = xcache_queue_alloc_line_data(xcache_io_queue(io), cache);
	if (dst == NULL) {
		return NULL;
	}

	ctx_data_cpy(cache->owner, dst, io->data, 0, from, size);
	return dst;
}

static int xcache_do_read_bf(struct xcache_backdev_io *base_io, uint64_t addr, uint64_t size)
{
	struct xcache_io *bf_io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(bf_io);
	ocf_cache_line_t line = base_io->line;

	int ret;
	uint64_t cache_addr;

	ret = xcache_read_bf_update_metadata(base_io, addr, size);
	if (ret != 0) {
		goto out;
	}

	cache_addr = cache_line_to_addr(cache, line, xcache_addr_offset(cache, addr));
	base_io->end = xcache_read_bf_cb;
	xcache_backdev_submit_io(base_io, true, cache_addr, size, 0, OCF_WRITE);
	return 0;

out:
	xcache_read_bf_done(cache, base_io);
	return ret;
}

static void xcache_read_bf_error(struct xcache_io *io, uint64_t addr, uint64_t size)
{
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_core_t core = xcache_io_core(io);
	ocf_core_id_t core_id = ocf_core_get_id(core);
	struct ocf_alock *alock = ocf_cache_line_concurrency(cache);

	uint64_t core_line_first, core_line_last, core_line;
	ocf_cache_line_t line;

	xcache_io_get_line_range(cache, addr, size, &core_line_first, &core_line_last);
	for (core_line = core_line_first; core_line <= core_line_last; core_line++) {
		line = xcache_engine_lookup_line(cache, core_id, core_line);
		if (line != INVALID_LINE) {
			xcache_unlock_wr(alock, line);
		}
	}

	ocf_core_stats_core_error_update(core, OCF_READ);
}

static void bf_io_end(struct xcache_io *bf_io, int error)
{
	xcache_queue_free_xcache_io(bf_io->io_queue, bf_io);
}

static struct xcache_io *xcache_get_bf_xcache_io(struct xcache_io *ori_io)
{
	struct xcache_io *bf_io = xcache_queue_alloc_xcache_io(ori_io->io_queue);

	if (bf_io == NULL) {
		ocf_cache_log(xcache_io_cache(ori_io), log_err, "alloc bf io failed\n");
		return NULL;
	}
	bf_io->io_queue = ori_io->io_queue;
	bf_io->cache = ori_io->cache;
	bf_io->core = ori_io->core;
	bf_io->error = 0;
	bf_io->end = bf_io_end;
	env_atomic_set(&bf_io->remaining, 1);
	return bf_io;
}

static void xcache_free_bf_xcache_io(struct xcache_io *bf_io)
{
	xcache_queue_free_xcache_io(bf_io->io_queue, bf_io);
}

static int xcache_submit_read_bf_line(struct xcache_io *io, struct xcache_io *bf_io, uint64_t bf_addr, uint64_t bf_size, ocf_cache_line_t line)
{
	ocf_cache_t cache = xcache_io_cache(bf_io);
	struct xcache_backdev_io *base_io = xcache_queue_alloc_backdev_io(xcache_io_queue(io));
	if (base_io == NULL) {
		ocf_cache_log(cache, log_err, "alloc bf base_io failed\n");
		xcache_unlock_wr(ocf_cache_line_concurrency(cache), line);
		return -1;
	}

	base_io->data = xcache_get_bf_data(io, bf_addr, bf_size);
	if (base_io->data == NULL) {
		ocf_cache_log(cache, log_err, "alloc bf_data failed\n");
		xcache_unlock_wr(ocf_cache_line_concurrency(cache), line);
		xcache_queue_free_backdev_io(xcache_io_queue(io), base_io);
		return -1;
	}
	base_io->xcache_io = bf_io;
	base_io->line = line;

	return xcache_do_read_bf(base_io, bf_addr, bf_size);
}

static void xcache_submit_read_bf(struct xcache_io *io, uint64_t addr, uint64_t size)
{
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_core_t core = xcache_io_core(io);
	ocf_core_id_t core_id = ocf_core_get_id(core);
	uint64_t line_size = ocf_line_size(cache);

	uint64_t core_line_first, core_line_last, core_line;
	ocf_cache_line_t line;
	uint64_t bf_addr, bf_size;
	struct xcache_backdev_io *base_io;
	struct xcache_io *bf_io;

	bf_io = xcache_get_bf_xcache_io(io);
	if (bf_io == NULL) {
		ocf_cache_log(cache, log_err, "alloc bf_io failed\n");
		xcache_read_bf_error(io, addr, size);
		return;
	}

	xcache_io_get_line_range(cache, addr, size, &core_line_first, &core_line_last);
	bf_addr = addr;
	bf_size = xcache_line_to_addr(cache, core_line_first + 1) - bf_addr;
	for (core_line = core_line_first; core_line <= core_line_last;
			core_line++, bf_addr += bf_size, bf_size = line_size) {
		if (core_line == core_line_last) {
			bf_size = (addr + size) - bf_addr;
		}

		line = xcache_engine_lookup_line(cache, core_id, core_line);
		if (line == INVALID_LINE) {
			continue;
		}

		if (xcache_submit_read_bf_line(io, bf_io, bf_addr, bf_size, line) != 0) {
			ocf_cache_log(cache, log_err, "read bf line failed\n");
			continue;
		}
	}

	xcache_io_put(bf_io);
}

static int xcache_read_bf(struct xcache_backdev_io *parent_base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = parent_base_io->xcache_io;

	xcache_queue_free_backdev_io(xcache_io_queue(parent_base_io->xcache_io), parent_base_io);

	if (cb_arg->error) {
		ocf_cache_log(xcache_io_cache(io), log_err, "read_miss failed with error %d\n",
				cb_arg->error);

		xcache_read_bf_error(io, cb_arg->addr, cb_arg->size);
		io->error = cb_arg->error;
		return cb_arg->error;
	}

	xcache_submit_read_bf(io, cb_arg->addr, cb_arg->size);

	return 0;
}

static int xcache_read_miss(struct xcache_backdev_io *base_io,
		uint64_t addr, uint64_t size, uint64_t buf_offset)
{
	base_io->end = xcache_read_bf;

	xcache_backdev_submit_io(base_io, false, addr, size, buf_offset, OCF_READ);
	return 0;
}

static int xcache_read_pt_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_core_t core = xcache_io_core(io);

	if (cb_arg->error) {
		ocf_cache_log(cache, log_err, "read pt failed\n");
		ocf_core_stats_core_error_update(core, OCF_READ);
	}

	xcache_queue_free_backdev_io(xcache_io_queue(base_io->xcache_io), base_io);
	return 0;
}

static int xcache_read_hit_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_core_t core = xcache_io_core(io);
	ocf_cache_line_t line = base_io->line;

	if (cb_arg->error) {
		ocf_cache_log(cache, log_err, "read hit failed\n");
		io->error = cb_arg->error;
		ocf_core_stats_cache_error_update(core, OCF_READ);
	}
	xcache_unlock_wr(ocf_cache_line_concurrency(cache), line);
	xcache_queue_free_backdev_io(xcache_io_queue(base_io->xcache_io), base_io);
	return 0;
}

static int xcache_read_hit(struct xcache_backdev_io *base_io, ocf_cache_line_t line,
		uint64_t offset, uint64_t size, uint64_t buf_offset)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	uint64_t addr = cache_line_to_addr(cache, line, offset);

	base_io->end = xcache_read_hit_cb;
	base_io->line = line;

	xcache_backdev_submit_io(base_io, true, addr, size, buf_offset, OCF_READ);
	return 0;
}

static void xcache_read_res(struct xcache_backdev_io *base_io)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	uint64_t start_addr = xcache_io_start_addr(io);

	uint8_t start_sector, last_sector;
	uint64_t buf_offset = base_io->addr - start_addr;

	start_sector = BYTES_TO_SECTORS(xcache_addr_offset(cache, base_io->addr));
	last_sector = start_sector + BYTES_TO_SECTORS(base_io->size - 1);

	if (metadata_test_valid_sec(cache, base_io->line, start_sector, last_sector)) {
		xcache_read_hit(base_io, base_io->line, SECTORS_TO_BYTES(start_sector),
				base_io->size, buf_offset);
	} else {
		xcache_read_miss(base_io, base_io->addr, base_io->size, buf_offset);
	}
}

static int xcache_read_line_valid(ocf_cache_t cache, ocf_cache_line_t line,
		uint8_t start_sector, uint8_t last_sector)
{
	return !metadata_test_valid_sec(cache, line, start_sector, last_sector);
}

static int xcache_read_lb_cb(struct xcache_backdev_io *backdev_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = backdev_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);

	xcache_unlock_wr(ocf_cache_line_concurrency(cache), backdev_io->line);
	xcache_queue_free_backdev_io(xcache_io_queue(backdev_io->xcache_io), backdev_io);
	return 0;
}

static void xcache_read_lb(struct xcache_backdev_io *backdev_io, ocf_cache_line_t line,
		uint64_t addr, uint64_t size)
{
	struct xcache_io *io = backdev_io->xcache_io;
	uint64_t start_addr = xcache_io_start_addr(io);
	uint64_t buf_offset = addr - start_addr;

	backdev_io->end = xcache_read_lb_cb;
	backdev_io->line = line;
	xcache_backdev_submit_io(backdev_io, false, addr, size, buf_offset, OCF_READ);
}

int xcache_read_generic(struct xcache_io *io)
{
	int ret;
	struct xcache_io_context ctx;
	struct xcache_io_handler read_handler = {
		.ctx = &ctx,
		.res_fn = xcache_read_res,
		.valid_fn = xcache_read_line_valid,
		.miss_fn = xcache_read_miss,
		.hit_fn = xcache_read_hit,
		.need_lb_fn = NULL,
		.lb_fn = xcache_read_lb,
	};

	xcache_init_io_ctx(&ctx, io);

	ret = xcache_handle_io(&read_handler);
	if (ret != 0) {
		io->error = ret;
		return ret;
	}

	xcache_update_stat(&ctx);
	return 0;
}
