#include "ocf/ocf.h"
#include "../ocf_cache_priv.h"
#include "../metadata/metadata.h"
#include "../utils/utils_cache_line.h"
#include "../concurrency/ocf_cache_line_concurrency.h"

#include "xcache_engine_common.h"
#include "xcache_engine_wb.h"
#include "xcache_engine_flush.h"
#include "../xcache_queue.h"
#include "../metadata/xcache_metadata.h"

static bool xcache_wb_update_metadata(ocf_cache_t cache, ocf_core_t core,
		ocf_cache_line_t line, uint8_t start_sector, uint8_t last_sector)
{
	ocf_part_id_t part_id = PARTITION_DEFAULT;
	struct ocf_part *part = &cache->user_parts[part_id].part;
	bool need_flush = false;

	bool line_was_dirty;

	ocf_metadata_start_collision_shared_access(cache, line);
	if (metadata_set_valid_sec_changed(cache, line, start_sector, last_sector)) {
		env_atomic_inc(&core->runtime_meta->cached_clines);
		env_atomic_inc(&core->runtime_meta->
				part_counters[part_id].cached_clines);
	}

	if (metadata_set_dirty_sec_changed(cache, line, start_sector, last_sector,
				&line_was_dirty)) {
		if (!line_was_dirty) {
			/*
			 * If this is first dirty cline set dirty timestamp
			 */
			if (!env_atomic64_read(&core->runtime_meta->dirty_since))
				env_atomic64_cmpxchg(
					&core->runtime_meta->dirty_since, 0,
					env_ticks_to_secs(env_get_tick_count()));

			/*
			 * Update the number of dirty cached data for that
			 * core object
			 */
			env_atomic_inc(&core->runtime_meta->dirty_clines);

			/*
			 * increment dirty clines statistic for given cline
			 */
			env_atomic_inc(&core->runtime_meta->
					part_counters[part_id].dirty_clines);

			need_flush = true;
		}
	}
	ocf_metadata_end_collision_shared_access(cache, line);
	return need_flush;
}

static void xcache_wb_flush_metadata_end(void *flush_io, int error)
{
	struct xcache_io *io = (struct xcache_io *)flush_io;
	ocf_cache_t cache = xcache_io_cache(io);
	struct ocf_alock *lock = ocf_cache_line_concurrency(cache);
	ocf_cache_line_t *flush_lines = xcache_io_get_flush_line(io);
	uint64_t i;

	for (i = 0; i < io->flush_line_num; i++) {
		xcache_unlock_wr(lock, flush_lines[i]);
	}
	xcache_io_free_flush_line(io);

	if (error) {
		ocf_cache_log(cache, log_err, "wb flush metadata failed with error %d\n", error);
		io->error = error;
	}

	xcache_io_put(io);
}

static int xcache_wb_hit_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	uint64_t addr = cb_arg->addr;
	uint64_t len = cb_arg->size;
	struct xcache_io *io = base_io->xcache_io;
	ocf_core_t core = xcache_io_core(io);
	ocf_cache_t cache = ocf_core_get_cache(core);
	int ret = 0;
	ocf_cache_line_t line = addr_to_cache_line(cache, addr);

	uint64_t offset;
	uint8_t start_sector, last_sector;

	xcache_queue_free_backdev_io(xcache_io_queue(base_io->xcache_io), base_io);

	if (cb_arg->error) {
		ocf_cache_log(cache, log_err, "wb write hit failed with error %d\n", cb_arg->error);
		ret = cb_arg->error;
		io->error = cb_arg->error;
		ocf_core_stats_cache_error_update(core, OCF_WRITE);
		goto out;
	}

	offset = xcache_addr_offset(cache, addr);
	start_sector = BYTES_TO_SECTORS(offset);
	last_sector = BYTES_TO_SECTORS(offset + len - 1);

	if (xcache_wb_update_metadata(cache, core, line, start_sector, last_sector)) {
		ret = xcache_io_add_flush_line(io, line);
		if (ret != 0) {
			ocf_cache_log(cache, log_err, "wb metadata add flush line failed with error %d\n", ret);
			io->error = ret;
			goto out;
		}
		if (xcache_metadata_should_flush(io)) {
			xcache_io_metadata_flush(io, xcache_wb_flush_metadata_end);
		}
		return 0;
	}

out:
	xcache_unlock_wr(ocf_cache_line_concurrency(cache), line);
	return ret;
}

static int xcache_wb_hit(struct xcache_backdev_io *base_io, ocf_cache_line_t line,
		uint64_t offset, uint64_t size, uint64_t buf_offset)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);

	uint64_t addr;

	addr = cache_line_to_addr(cache, line, offset);
	base_io->end = xcache_wb_hit_cb;
	xcache_backdev_submit_io(base_io, true, addr, size, buf_offset, OCF_WRITE);
	return 0;
}

static int xcache_wb_miss_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_core_t core = xcache_io_core(io);

	if (cb_arg->error != 0) {
		ocf_cache_log(cache, log_err, "wb miss with error %d\n", cb_arg->error);
		io->error = cb_arg->error;
		ocf_core_stats_core_error_update(core, OCF_WRITE);
	}
	xcache_queue_free_backdev_io(xcache_io_queue(base_io->xcache_io), base_io);
	return 0;
}

static int xcache_wb_miss(struct xcache_backdev_io *base_io, uint64_t addr,
		uint64_t size, uint64_t buf_offset)
{
	base_io->end = xcache_wb_miss_cb;
	xcache_backdev_submit_io(base_io, false, addr, size, buf_offset, OCF_WRITE);
	return 0;
}

static void xcache_wb_res(struct xcache_backdev_io *base_io)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	uint64_t offset = xcache_addr_offset(cache, base_io->addr);
	uint64_t start_addr = xcache_io_start_addr(io);
	uint64_t buf_offset = base_io->addr - start_addr;

	xcache_wb_hit(base_io, base_io->line, offset, base_io->size, buf_offset);
}

/* bypass dirty sectors to core will make additional cache io to update valid bit */
static int xcache_wb_need_lb(ocf_cache_t cache, struct xcache_line_range *range)
{
	return !metadata_test_dirty_sec(cache, range->cache_line, range->start_sector, range->last_sector);
}

static void xcache_wb_lb(struct xcache_backdev_io *backdev_io, ocf_cache_line_t line,
		uint64_t addr, uint64_t size)
{
	xcache_wr_lb_common(backdev_io, line, addr, size);
}

int xcache_wb(struct xcache_io *io)
{
	int ret;
	struct xcache_io_context ctx;
	struct xcache_io_handler wb_handler = {
		.ctx = &ctx,
		.res_fn = xcache_wb_res,
		.valid_fn = NULL,
		.miss_fn = xcache_wb_miss,
		.hit_fn = xcache_wb_hit,
		.need_lb_fn = xcache_wb_need_lb,
		.lb_fn = xcache_wb_lb,
	};

	mark_flush();

	io->flush_line_num = 0;
	xcache_init_io_ctx(&ctx, io);

	ret = xcache_handle_io(&wb_handler);
	if (ret != 0) {
		return ret;
	}

	xcache_update_stat(&ctx);
	return 0;
}
