#include "ocf/ocf.h"
#include "../ocf_cache_priv.h"
#include "../utils/utils_cache_line.h"
#include "../metadata/metadata.h"

#include "xcache_engine_wt.h"
#include "xcache_engine_common.h"
#include "xcache_engine_flush.h"
#include "../xcache_queue.h"

static void xcache_wt_udpate_metadata(ocf_cache_t cache, ocf_core_t core,
		ocf_cache_line_t line, uint8_t start_sector, uint8_t last_sector)
{
	ocf_part_id_t part_id = PARTITION_DEFAULT;
	struct ocf_part *part = &cache->user_parts[part_id].part;

	bool line_is_clean;

	ocf_metadata_start_collision_shared_access(cache, line);
	if (metadata_set_valid_sec_changed(cache, line, start_sector, last_sector)) {
		env_atomic_inc(&core->runtime_meta->cached_clines);
		env_atomic_inc(&core->runtime_meta->
				part_counters[part_id].cached_clines);
	}

	if (metadata_clear_dirty_sec_changed(cache, line, start_sector, last_sector,
				&line_is_clean)) {
		if (line_is_clean) {
			/*
			 * Update the number of dirty cached data for that
			 * core object
			 */
			if (env_atomic_dec_and_test(&core->runtime_meta->
						dirty_clines)) {
				/*
				 * If this is last dirty cline reset dirty
				 * timestamp
				 */
				env_atomic64_set(&core->runtime_meta->
						dirty_since, 0);
			}

			/*
			 * decrement dirty clines statistic for given cline
			 */
			env_atomic_dec(&core->runtime_meta->
					part_counters[part_id].dirty_clines);
		}
	}
	ocf_metadata_end_collision_shared_access(cache, line);
}

static int xcache_wt_hit_core_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_core_t core = xcache_io_core(io);
	ocf_cache_t cache = ocf_core_get_cache(core);
	int ret = 0;


	if (cb_arg->error) {
		ret = cb_arg->error;
		ocf_cache_log(cache, log_err, "wt hit core failed with error %d\n", cb_arg->error);
		io->error = cb_arg->error;
		ocf_core_stats_core_error_update(core, OCF_WRITE);
	}

	xcache_queue_free_backdev_io(xcache_io_queue(io), base_io);
	return ret;
}

static int xcache_wt_hit_cache_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	uint64_t addr = cb_arg->addr;
	uint64_t len = cb_arg->size;
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_core_t core = xcache_io_core(io);
	ocf_cache_line_t line = base_io->line;
	int ret = 0;

	uint64_t offset;
	uint8_t start_sector, last_sector;

	if (cb_arg->error != 0) {
		ocf_cache_log(cache, log_err, "wt hit cache with error %d\n", cb_arg->error);
		io->error = cb_arg->error;
		ocf_core_stats_cache_error_update(core, OCF_WRITE);
		ret = cb_arg->error;
		goto out;
	}

	offset = xcache_addr_offset(cache, addr);
	start_sector = BYTES_TO_SECTORS(offset);
	last_sector = BYTES_TO_SECTORS(offset + len - 1);

	xcache_wt_udpate_metadata(cache, core, line, start_sector, last_sector);

out:
	xcache_unlock_wr(ocf_cache_line_concurrency(cache), line);
	xcache_queue_free_backdev_io(xcache_io_queue(base_io->xcache_io), base_io);
	return ret;
}

static int xcache_wt_hit_cache(struct xcache_backdev_io *base_io, ocf_cache_line_t line,
		uint64_t offset, uint64_t size, uint64_t buf_offset)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	uint64_t addr = cache_line_to_addr(cache, line, offset);

	base_io->line = line;
	base_io->end = xcache_wt_hit_cache_cb;
	xcache_backdev_submit_io(base_io, true, addr, size, buf_offset, OCF_WRITE);
	return 0;
}

static inline int xcache_wt_hit(struct xcache_backdev_io *base_io, ocf_cache_line_t line,
		uint64_t addr, uint64_t offset, uint64_t size, uint64_t buf_offset)
{
	return xcache_wt_hit_cache(base_io, line, offset, size, buf_offset);
}

static void xcache_wt_res(struct xcache_backdev_io *base_io)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	uint64_t offset = xcache_addr_offset(cache, base_io->addr);
	uint64_t start_addr = xcache_io_start_addr(io);
	uint64_t buf_offset = base_io->addr - start_addr;

	xcache_wt_hit(base_io, base_io->line, base_io->addr, offset, base_io->size, buf_offset);
}

static int xcache_wt_miss_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_io *io = base_io->xcache_io;
	ocf_cache_t cache = xcache_io_cache(io);
	ocf_core_t core = xcache_io_core(io);

	if (cb_arg->error != 0) {
		ocf_cache_log(cache, log_err, "wt miss with error %d\n", cb_arg->error);
		io->error = cb_arg->error;
		ocf_core_stats_core_error_update(core, OCF_WRITE);
	}
	xcache_queue_free_backdev_io(xcache_io_queue(base_io->xcache_io), base_io);
	return 0;
}

static int xcache_wt_core(struct xcache_io_context *ctx)
{
	struct xcache_io *io = ctx->io;
	uint64_t addr = xcache_io_start_addr(io);
	uint64_t size = xcache_io_size(io);
	ocf_cache_t cache = xcache_io_cache(io);
	struct xcache_backdev_io *base_io = NULL;

	base_io = xcache_queue_alloc_backdev_io(xcache_io_queue(io));
	if (base_io == NULL) {
		ocf_cache_log(cache, log_err, "alloc base io failed\n");
		return -ENOMEM;
	}
	base_io->xcache_io = io;
	base_io->data = io->data;
	base_io->end = xcache_wt_miss_cb;
	xcache_backdev_submit_io(base_io, false, addr, size, 0, OCF_WRITE);
	return 0;
}

static void xcache_wt_lb(struct xcache_backdev_io *backdev_io, ocf_cache_line_t line,
		uint64_t addr, uint64_t size)
{
	xcache_wr_lb_common(backdev_io, line, addr, size);
}

int xcache_wt(struct xcache_io *io)
{
	struct xcache_io_context ctx;
	struct xcache_io_handler wt_handler = {
		.ctx = &ctx,
		.res_fn = xcache_wt_res,
		.valid_fn = NULL,
		.miss_fn = NULL,
		.hit_fn = xcache_wt_hit_cache,
		.need_lb_fn = NULL,
		.lb_fn = xcache_wt_lb,
	};
	int ret;

	mark_flush();

	xcache_init_io_ctx(&ctx, io);

	if (xcache_wt_core(&ctx) != 0) {
		io->error = -ENOMEM;
		return -ENOMEM;
	}

	ret = xcache_handle_io(&wt_handler);
	if (ret != 0) {
		io->error = ret;
		return ret;
	}

	ctx.core_bytes = xcache_io_size(io);
	xcache_update_stat(&ctx);
	return 0;
}
