#include <stdint.h>

#include "deadline.h"
#include "evicting_ops.h"
#include "evicting_helper.h"
#include "../xcache.h"

#define STALE_SECS 600
#define EVICT_BUNCH 10
#define CLEAN_BUNCH 32

struct deadline_ctrl {
	uint32_t *access_tick;
	uint64_t line_num;
	xcache_line_t evict_iter_line;
	xcache_line_t clean_iter_line;
};

struct deadline_iter {
	xcache_line_t *lines;
	int line_num;
	int iter_bunch;
	xcache_line_t *iter_start;
	bool line_dirty;
};

static int deadline_init(xcache_context_t *ctx)
{
	struct deadline_ctrl *ctrl = env_malloc(sizeof(struct deadline_ctrl), 0);

	if (ctrl == NULL) {
		return -1;
	}
	ctrl->line_num = evicting_line_num(ctx);

	ctrl->access_tick = env_zalloc(sizeof(uint32_t) * ctrl->line_num, 0);
	if (ctrl->access_tick == NULL) {
		goto alloc_tick_fail;
	}

	ctrl->evict_iter_line = 0;
	ctrl->clean_iter_line = 0;
	xcache_set_evicting(ctx, (void *)ctrl);
	return 0;

alloc_tick_fail:
	env_free(ctrl);
	return -1;
}

static void deadline_fini(xcache_context_t *ctx)
{
	struct deadline_ctrl *ctrl = (struct deadline_ctrl *)xcache_get_evicting(ctx);

	if (ctrl == NULL) {
		return;
	}

	xcache_set_evicting(ctx, NULL);

	if (ctrl->access_tick != NULL) {
		env_free(ctrl->access_tick);
		ctrl->access_tick = NULL;
	}

	env_free(ctrl);
}

static void deadline_line_accessed(xcache_context_t *ctx, xcache_line_t line)
{
	struct deadline_ctrl *ctrl = (struct deadline_ctrl *)xcache_get_evicting(ctx);

	ctrl->access_tick[line] = env_ticks_to_secs(env_get_tick_count());
}

static void deadline_line_dirty(xcache_context_t *ctx, xcache_line_t line)
{
	return;
}

static void deadline_line_clean(xcache_context_t *ctx, xcache_line_t line)
{
	return;
}

static int deadline_get_lines(xcache_context_t *ctx, struct deadline_iter *iter)
{
	struct deadline_ctrl *ctrl = xcache_get_evicting(ctx);
	uint32_t stale = env_ticks_to_secs(env_get_tick_count()) - STALE_SECS;
	uint64_t i, j;
	xcache_line_t cline;

	j = 0;
	for (i = 0, cline = *(iter->iter_start);
			i < iter->iter_bunch;
			i++, cline = (cline + 1) % ctrl->line_num)
	{
		if (ctrl->access_tick[cline] > stale) {
			continue;
		}

		if (!evicting_trylock_line(ctx, cline)) {
			continue;
		}

		if (evicting_test_dirty(ctx, cline) != iter->line_dirty) {
			evicting_unlock_line(ctx, cline);
			continue;
		}

		iter->lines[j++] = cline;
		if (j >= iter->line_num) {
			*(iter->iter_start) = (cline + 1) % ctrl->line_num;
			return j;
		}
	}

	*(iter->iter_start) = cline;
	return j;
}

static int deadline_line_to_clean(xcache_context_t *ctx, xcache_line_t *lines, int line_num)
{
	struct deadline_ctrl *ctrl = xcache_get_evicting(ctx);

	struct deadline_iter iter = {
		.lines = lines,
		.line_num = line_num,
		.iter_bunch = CLEAN_BUNCH,
		.iter_start = &(ctrl->clean_iter_line),
		.line_dirty = true,
	};
	return deadline_get_lines(ctx, &iter);
}


static int deadline_line_to_evict(xcache_context_t *ctx, xcache_line_t *lines, int line_num)
{
	struct deadline_ctrl *ctrl = xcache_get_evicting(ctx);

	struct deadline_iter iter = {
		.lines = lines,
		.line_num = line_num,
		.iter_bunch = EVICT_BUNCH,
		.iter_start = &(ctrl->evict_iter_line),
		.line_dirty = false,
	};
	return deadline_get_lines(ctx, &iter);
}

static void deadline_destroy(xcache_context_t *ctx)
{
	struct deadline_ctrl *ctrl = (struct deadline_ctrl *)ctx->xcache_evicting;

	env_free(ctrl->access_tick);
	env_free(ctrl);
}

static struct evicting_policy_ops g_deadline_ops = {
	.init = deadline_init,
	.fini = deadline_fini,
	.line_accessed = deadline_line_accessed,
	.line_dirty = deadline_line_dirty,
	.line_clean = deadline_line_clean,
	.line_to_clean = deadline_line_to_clean,
	.line_to_evict = deadline_line_to_evict,
};

void set_deadline_policy(void)
{
	evicting_set(&g_deadline_ops);
}
