#ifndef __QOS_LB_H__
#define __QOS_LB_H__

#define CORE_DEFAULT_LOAD_WEIGHT 30
#define CACHE_DEFAULT_LOAD_WEIGHT 1

#define CORE_DEV 0
#define CACHE_DEV 1

#include <stdint.h>
#include <stdbool.h>

#include "ocf/xcache.h"

struct qos_dev_load {
	env_atomic64 read_inflight_bytes;
	env_atomic64 write_inflight_bytes;
	uint32_t read_weight;
	uint32_t write_weight;
};

static inline void qos_dev_load_init(struct qos_dev_load *load)
{
	env_atomic64_set(&load->read_inflight_bytes, 0);
	env_atomic64_set(&load->write_inflight_bytes, 0);
}

static inline void qos_dev_load_add(struct qos_dev_load *load, enum xcache_dir dir, uint64_t bytes)
{
	switch (dir) {
		case XCACHE_RD:
			env_atomic64_add(bytes, &load->read_inflight_bytes);
			break;
		case XCACHE_WR:
			env_atomic64_add(bytes, &load->write_inflight_bytes);
			break;
		default:
			break;
	}
}

static inline void qos_dev_load_sub(struct qos_dev_load *load, enum xcache_dir dir, uint64_t bytes)
{
	switch (dir) {
		case XCACHE_RD:
			env_atomic64_sub(bytes, &load->read_inflight_bytes);
			break;
		case XCACHE_WR:
			env_atomic64_sub(bytes, &load->write_inflight_bytes);
			break;
		default:
			break;
	}
}

static inline uint64_t qos_dev_load_read(struct qos_dev_load *load, enum xcache_dir dir)
{
	switch (dir) {
		case XCACHE_RD:
			return env_atomic64_read(&load->read_inflight_bytes);
		case XCACHE_WR:
			return env_atomic64_read(&load->write_inflight_bytes);
		default:
			return 0;
	}
}

static inline uint32_t qos_dev_load_weight(struct qos_dev_load *load, enum xcache_dir dir)
{
	switch (dir) {
		case XCACHE_RD:
			return load->read_weight;
		case XCACHE_WR:
			return load->write_weight;
		default:
			return 0;
	}
}

static inline uint64_t do_cal_load(uint64_t bytes, uint32_t weight)
{
	return bytes * weight;
}

static inline uint64_t qos_dev_load_cal(struct qos_dev_load *load)
{
	uint64_t read_inflight_bytes = qos_dev_load_read(load, XCACHE_RD);
	uint64_t write_inflight_bytes = qos_dev_load_read(load, XCACHE_WR);

	return do_cal_load(read_inflight_bytes, load->read_weight) +
		do_cal_load(write_inflight_bytes, load->write_weight);
}

struct qos_lb {
	struct qos_dev_load cache_load;
	struct qos_dev_load core_load;
};

static inline void qos_lb_init(struct qos_lb *qos_lb)
{
	qos_dev_load_init(&qos_lb->cache_load);
	qos_dev_load_init(&qos_lb->core_load);
	qos_lb->cache_load.read_weight = CACHE_DEFAULT_LOAD_WEIGHT;
	qos_lb->cache_load.write_weight = CACHE_DEFAULT_LOAD_WEIGHT;
	qos_lb->core_load.read_weight = CORE_DEFAULT_LOAD_WEIGHT;
	qos_lb->core_load.write_weight = CORE_DEFAULT_LOAD_WEIGHT;
}

static inline void qos_lb_load_add(struct qos_lb *qos_lb, int dev, enum xcache_dir dir, uint64_t bytes)
{
	switch (dev) {
		case CACHE_DEV:
			qos_dev_load_add(&qos_lb->cache_load, dir, bytes);
			break;
		case CORE_DEV:
			qos_dev_load_add(&qos_lb->core_load, dir, bytes);
			break;
		default:
			break;
	}
}

static inline void qos_lb_load_sub(struct qos_lb *qos_lb, int dev, enum xcache_dir dir, uint64_t bytes)
{
	switch (dev) {
		case CACHE_DEV:
			qos_dev_load_sub(&qos_lb->cache_load, dir, bytes);
			break;
		case CORE_DEV:
			qos_dev_load_sub(&qos_lb->core_load, dir, bytes);
			break;
		default:
			break;
	}
}

static inline bool qos_need_lb(struct qos_lb *qos_lb, enum xcache_dir dir, uint64_t bytes)
{
	return qos_dev_load_cal(&qos_lb->cache_load) > qos_dev_load_cal(&qos_lb->core_load) +
		do_cal_load(bytes, qos_dev_load_weight(&qos_lb->core_load, dir));
}

#endif
