#include "metadata/metadata.h"
#include "concurrency/ocf_concurrency.h"
#include "utils/utils_cleaner.h"
#include "utils/utils_cache_line.h"
#include "ocf_queue_priv.h"
#include "cleaning/cleaning.h"

#include "ocf/xcache.h"

#include "xcache_cleaner.h"
#include "xcache.h"
#include "engine/xcache_engine_common.h"
#include "xcache_queue.h"
#include "evicting/evicting_ops.h"
#include "metadata/xcache_metadata.h"

static inline uint8_t get_first_sector(ocf_cache_t cache, ocf_cache_line_t line, uint8_t start_sector, bool dirty)
{
	uint8_t sector;

	for (sector = start_sector; sector < ocf_line_sectors(cache); sector++) {
		if (metadata_test_dirty_one(cache, line, sector) == dirty) {
			break;
		}
	}
	return sector;
}

static inline uint8_t get_first_clean_sector(ocf_cache_t cache, ocf_cache_line_t line, uint8_t start_sector)
{
	return get_first_sector(cache, line, start_sector, false);
}

static inline uint8_t get_first_dirty_sector(ocf_cache_t cache, ocf_cache_line_t line, uint8_t start_sector)
{
	return get_first_sector(cache, line, start_sector, true);
}

static inline void get_dirty_sectors(ocf_cache_t cache, ocf_cache_line_t line,
		uint8_t *start_sector, uint8_t *end_sector)
{
	bool dirty;

	uint8_t sector = *start_sector;
	uint8_t line_sectors = ocf_line_sectors(cache);

	// fast path
	if (*start_sector == 0 && metadata_test_dirty_all(cache, line)) {
		*end_sector = line_sectors - 1;
		return;
	}

	*start_sector = get_first_dirty_sector(cache, line, *start_sector);
	if (*start_sector >= line_sectors) {
		return;
	}

	*end_sector = get_first_clean_sector(cache, line, *start_sector + 1) - 1;
}

static inline void xcache_clean_unlock_line(struct xcache_cleaning_ctx *cleaning_ctx)
{
	struct clean_sort_data *data = cleaning_ctx->data;
	struct ocf_alock *c = ocf_cache_line_concurrency(cleaning_ctx->cache);
	int i;

	for (i = 0; i < cleaning_ctx->count; i++) {
		xcache_unlock_wr(c, data[i].line);
	}
}

static inline void xcache_clean_cleanup(struct xcache_cleaning_ctx *cleaning_ctx)
{
	struct xcache_backdev_io *base_io = &cleaning_ctx->base_io;
	ocf_cache_t cache = cleaning_ctx->cache;

	int i;

	if (base_io->data != NULL) {
		ctx_data_free(cache->owner, base_io->data);
		base_io->data = NULL;
	}

	xcache_cleaner_complete(cleaning_ctx);
}

static inline void xcache_clean_cb(struct xcache_cleaning_ctx *cleaning_ctx)
{
	if (cleaning_ctx->end) {
		cleaning_ctx->end(cleaning_ctx->arg);
		cleaning_ctx->end = NULL;
	}
}

static void xcache_clean_end(struct xcache_cleaning_ctx *cleaning_ctx)
{
	xcache_clean_unlock_line(cleaning_ctx);
	xcache_clean_cleanup(cleaning_ctx);
	xcache_clean_cb(cleaning_ctx);
}

static void xcache_clean_update_metadata(ocf_cache_t cache, ocf_core_t core,
		ocf_cache_line_t line)
{
	ocf_part_id_t part_id = PARTITION_DEFAULT;
	struct ocf_part *part = &cache->user_parts[part_id].part;

	ocf_metadata_start_collision_shared_access(cache, line);
	metadata_clear_dirty(cache, line);

	if (env_atomic_dec_and_test(&core->runtime_meta->
				dirty_clines)) {
		env_atomic64_set(&core->runtime_meta->
				dirty_since, 0);
	}

	env_atomic_dec(&core->runtime_meta->
			part_counters[part_id].dirty_clines);
	ocf_metadata_end_collision_shared_access(cache,
			line);
}

static int clean_line_num(void *clean_io)
{
	struct xcache_cleaning_ctx *cleaning_ctx = (struct xcache_cleaning_ctx *)clean_io;

	return cleaning_ctx->count;
}

static bool clean_flush_line(void *clean_io, int index, ocf_cache_line_t *line)
{
	struct xcache_cleaning_ctx *cleaning_ctx = (struct xcache_cleaning_ctx *)clean_io;

	*line = cleaning_ctx->data[index].line;
	return true;
}

static void xcache_clean_flush_metadata_end(void *clean_io, int error)
{
	struct xcache_cleaning_ctx *cleaning_ctx = (struct xcache_cleaning_ctx *)clean_io;

	xcache_clean_end(cleaning_ctx);
}

static void xcache_clean_metadata_flush(struct xcache_cleaning_ctx *cleaning_ctx)
{
	struct xcache_io *io = &cleaning_ctx->io;
	struct ocf_metadata_io_context io_ctx = {
		.io = (void *)cleaning_ctx,
		.io_flags = 0,
		.io_end = xcache_clean_flush_metadata_end,
		.queue = xcache_io_queue(io),
	};
	struct ocf_metadata_line_getter line_getter = {
		.getter = (void *)cleaning_ctx,
		.get_line_num = clean_line_num,
		.get_flush_line = clean_flush_line,
	};

	xcache_metadata_flush_asynch(cleaning_ctx->cache, &io_ctx, &line_getter);
}

static void xcache_clean_bf_end(struct xcache_io *io, int error)
{
	struct xcache_cleaning_ctx *cleaning_ctx = container_of(io, struct xcache_cleaning_ctx,
			io);
	struct clean_sort_data *data = cleaning_ctx->data;
	ocf_cache_t cache = cleaning_ctx->cache;
	struct ocf_alock *c = ocf_cache_line_concurrency(cleaning_ctx->cache);

	ocf_core_t core;
	ocf_cache_line_t line;
	int i, j;

	if (error != 0) {
		xcache_clean_end(cleaning_ctx);
		return;
	}

	for (i = 0, j = 0; i < cleaning_ctx->count; i++) {
		// core_line write failed
		if (data[i].core_line == ULLONG_MAX) {
			xcache_unlock_wr(c, data[i].line);
			continue;
		}

		core = ocf_cache_get_core(cache, data[i].core_id);
		line = data[i].line;
		xcache_clean_update_metadata(cache, core, line);
		// record line to flush
		data[j++].line = data[i].line;
	}
	cleaning_ctx->count = j;

	if (j == 0) {
		xcache_clean_end(cleaning_ctx);
	} else {
		xcache_clean_metadata_flush(cleaning_ctx);
	}
}

static void remove_failed_line(struct xcache_cleaning_ctx *cleaning_ctx)
{
	int l, r;

	l = 0;
	for (r = 0; r < cleaning_ctx->count; r++) {
		if (cleaning_ctx->data[r].line == INVALID_LINE) {
			continue;
		}
		if (l != r) {
			cleaning_ctx->data[l] = cleaning_ctx->data[r];
		}
		l++;
	}
	cleaning_ctx->count = l;
}

static int xcache_clean_write_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_cleaning_ctx *cleaning_ctx = container_of(base_io,
			struct xcache_cleaning_ctx, base_io);
	struct clean_sort_data *data = cleaning_ctx->data;
	ocf_cache_t cache = cleaning_ctx->cache;

	uint64_t first_line, last_line;
	int i;
	ocf_core_t core;

	if (cb_arg->error == 0) {
		return 0;
	}

	// core_line in data array is sorted in xcache_cleaner_clean
	first_line = xcache_addr_to_line(cache, cb_arg->addr);
	last_line = xcache_addr_to_line(cache, cb_arg->addr + cb_arg->size - 1);
	for (i = 0; i < cleaning_ctx->count; i++) {
		if (data[i].core_line >= first_line) {
			break;
		}
	}

	for (; i < cleaning_ctx->count; i++) {
		if (data[i].core_line > last_line) {
			break;
		}
		// cache line alock is still hold, don't touch data[i].line
		data[i].core_line = ULLONG_MAX;
		core = ocf_cache_get_core(cache, data[i].core_id);
		ocf_core_stats_core_error_update(core, OCF_WRITE);
	}

	return 0;
}

struct dirty_range {
	uint8_t start_sector;
	uint8_t last_sector;
	int idx;
	uint64_t core_line;
};

struct clear_range {
	uint8_t start_sector;
	uint8_t last_sector;
	int start_idx;
	uint64_t start_line;
	uint64_t last_line;
};

static void get_dirty_range(ocf_cache_t cache, struct xcache_cleaning_ctx *cleaning_ctx, struct dirty_range *range)
{
	struct clean_sort_data *data = cleaning_ctx->data;

	for (; range->idx < cleaning_ctx->count; range->idx++) {
		get_dirty_sectors(cache, data[range->idx].line, &range->start_sector, &range->last_sector);
		if (range->start_sector < ocf_line_sectors(cache)) {
			range->core_line = data[range->idx].core_line;
			break;
		}

		range->start_sector = 0;
	}
}

static void dirty_range_next(ocf_cache_t cache, struct dirty_range *range)
{
	if (range->start_sector >= ocf_line_sectors(cache)) {
		range->start_sector = 0;
		range->idx++;
	} else {
		range->start_sector++;
	}
}

static int merge_dirty_range(ocf_cache_t cache, struct clear_range *clear_range, struct dirty_range *dirty_range)
{
	if (clear_range->last_sector != ocf_line_sectors(cache) - 1) {
		return -1;
	}

	if (dirty_range->start_sector != 0) {
		return -1;
	}

	if (dirty_range->core_line != clear_range->last_line + 1) {
		return -1;
	}

	clear_range->last_line = dirty_range->core_line;
	clear_range->last_sector = dirty_range->last_sector;
	return 0;
}

static void submit_dirty_range(ocf_cache_t cache, struct xcache_cleaning_ctx *cleaning_ctx, struct clear_range *clear_range)
{
	struct xcache_backdev_io *base_io = &cleaning_ctx->base_io;
	struct xcache_io *io = &cleaning_ctx->io;
	uint64_t addr, size, buf_offset;

	addr = xcache_line_to_addr(cache, clear_range->start_line) + SECTORS_TO_BYTES(clear_range->start_sector);
	size = xcache_line_to_addr(cache, clear_range->last_line - clear_range->start_line) + SECTORS_TO_BYTES((int)clear_range->last_sector - (int)clear_range->start_sector + 1);
	buf_offset = xcache_line_to_addr(cache, clear_range->start_idx) + SECTORS_TO_BYTES(clear_range->start_sector);
	xcache_backdev_submit_io(base_io, false, addr, size, buf_offset, OCF_WRITE);
}

static void xcache_clean_bf(struct xcache_io *io, int error)
{
	struct xcache_cleaning_ctx *cleaning_ctx = container_of(io,
			struct xcache_cleaning_ctx, io);
	struct xcache_backdev_io *base_io = &cleaning_ctx->base_io;
	struct clean_sort_data *data = cleaning_ctx->data;
	ocf_cache_t cache = cleaning_ctx->cache;

	struct dirty_range dirty_range;
	struct clear_range clear_range;

	remove_failed_line(cleaning_ctx);

	if (error) {
		xcache_clean_bf_end(io, error);
	}

	env_atomic_set(&io->remaining, 1);
	io->end = xcache_clean_bf_end;
	base_io->end = xcache_clean_write_cb;

	dirty_range.start_sector = 0;
	dirty_range.idx = 0;
	get_dirty_range(cache, cleaning_ctx, &dirty_range);
	while (dirty_range.idx < cleaning_ctx->count) {
		clear_range.start_sector = dirty_range.start_sector;
		clear_range.start_line = data[dirty_range.idx].core_line;
		clear_range.start_idx = dirty_range.idx;
		clear_range.last_sector = dirty_range.last_sector;
		clear_range.last_line = data[dirty_range.idx].core_line;

		for (dirty_range_next(cache, &dirty_range);
				dirty_range.idx < cleaning_ctx->count;
				dirty_range_next(cache, &dirty_range)) {
			get_dirty_range(cache, cleaning_ctx, &dirty_range);
			if (merge_dirty_range(cache, &clear_range, &dirty_range) != 0) {
				break;
			}
		}
		submit_dirty_range(cache, cleaning_ctx, &clear_range);
	}

	xcache_io_put(io);
}

static int xcache_clean_read_cb(struct xcache_backdev_io *base_io, struct backdev_io_end_arg *cb_arg)
{
	struct xcache_cleaning_ctx *cleaning_ctx = container_of(base_io,
			struct xcache_cleaning_ctx, base_io);
	struct ocf_alock *c = ocf_cache_line_concurrency(cleaning_ctx->cache);

	ocf_cache_line_t line;
	ocf_core_t core;
	int i;

	if (cb_arg->error == 0) {
		return 0;
	}

	line = addr_to_cache_line(cleaning_ctx->cache, cb_arg->addr);
	for (i = 0; i < cleaning_ctx->count; i++) {
		if (cleaning_ctx->data[i].line == line) {
			xcache_unlock_wr(c, line);
			cleaning_ctx->data[i].line = INVALID_LINE;
			core = ocf_cache_get_core(cleaning_ctx->cache, cleaning_ctx->data[i].core_id);
			ocf_core_stats_cache_error_update(core, OCF_READ);
			break;
		}
	}

	return 0;
}

static int xcache_clean_if(struct xcache_io *io)
{
	struct xcache_cleaning_ctx *cleaning_ctx = container_of(io,
			struct xcache_cleaning_ctx, io);
	struct xcache_backdev_io *base_io = &cleaning_ctx->base_io;
	ocf_cache_t cache = cleaning_ctx->cache;
	uint64_t line_size = ocf_line_size(cache);
	uint64_t buf_offset = 0;

	ocf_cache_line_t line;
	uint64_t addr;
	int i;

	io->end = xcache_clean_bf;
	base_io->end = xcache_clean_read_cb;
	for (i = 0; i < cleaning_ctx->count; i++) {
		line = cleaning_ctx->data[i].line;
		addr = cache_line_to_addr(cache, line, 0);
		xcache_backdev_submit_io(base_io, true, addr, line_size, buf_offset, OCF_READ);
		buf_offset += line_size;
	}
	return 0;
}

static inline int sort_data_cmp(struct clean_sort_data *l, struct clean_sort_data *r)
{
	if (l->core_id != r->core_id) {
		return l->core_id - r->core_id;
	}
	return l->core_line - r->core_line;
}

static inline void swap_sort_data(struct clean_sort_data *l, struct clean_sort_data *r)
{
	struct clean_sort_data tmp;

	tmp = *l;
	*l = *r;
	*r = tmp;
}

static void clean_quick_sort(struct clean_sort_data *data, int start, int end)
{
	int i, j;

	if (start >= end) {
		return;
	}

	j = start;
	for (i = start; i < end; i++) {
		if (sort_data_cmp(&data[i], &data[end]) < 0) {
			if (i != j) {
				swap_sort_data(&data[i], &data[j]);
			}
			j++;
		}
	}
	if (j != end) {
		swap_sort_data(&data[j], &data[end]);
	}

	clean_quick_sort(data, start, j - 1);
	clean_quick_sort(data, j + 1, end);
}

static int xcache_clean_sort(ocf_cache_t cache, struct xcache_cleaning_ctx *cleaning_ctx)
{
	struct clean_sort_data *data = cleaning_ctx->data;
	int count = cleaning_ctx->count;
	int i;

	for (i = 0; i < count; i++) {
		ocf_metadata_get_core_info(cache, data[i].line, &data[i].core_id,
				&data[i].core_line);
	}
	clean_quick_sort(data, 0, count - 1);

	return 0;
}

/*
 * 1. we can't use spdk_io_to_* here
 * 2. xcache_cleaner_prepare must be called
 */
int xcache_cleaner_clean(ocf_cache_t cache, ocf_queue_t q,
		struct xcache_cleaning_ctx *cleaning_ctx)
{
	struct xcache_io *io = &cleaning_ctx->io;
	struct xcache_backdev_io *base_io = &cleaning_ctx->base_io;

	cleaning_ctx->cache = cache;

	if (cleaning_ctx->count == 0) {
		goto err_out;
	}

	if (xcache_clean_sort(cache, cleaning_ctx) != 0) {
		goto err_out;
	}

	base_io->data = ctx_data_alloc(cache->owner, ocf_line_pages(cache) * cleaning_ctx->count);
	if (base_io->data == NULL) {
		goto err_out;
	}

	base_io->xcache_io = io;
	base_io->end = NULL;

	io->io_if = xcache_clean_if;
	io->io_queue = q;
	io->error = 0;
	env_atomic_set(&io->remaining, 1);
	xcache_queue_push_xcache_io_back(io, true);
	return 0;

err_out:
	xcache_clean_end(cleaning_ctx);
	return -1;
}

#define CLEAN_INTERVAL 600000 // 600 seconds

static void xcache_cleaner_run_complete(void *arg)
{
	ocf_cleaner_t cleaner = (ocf_cleaner_t)arg;

	ocf_cleaner_run_complete(cleaner, CLEAN_INTERVAL);
}

void xcache_cleaner_run(ocf_cleaner_t cleaner, ocf_queue_t queue)
{
	ocf_cache_t cache;
	struct xcache_cleaning_ctx *cleaning_ctx;
	xcache_context_t *xcache_ctx;
	ocf_cache_line_t cline[XCACHE_CLEAN_SIZE];
	int cline_num;
	unsigned lock_idx;

	if (cleaner == NULL || queue == NULL) {
		return;
	}

	cache = ocf_cleaner_get_cache(cleaner);
	cleaning_ctx = xcache_get_cleaning_ctx(cache);
	xcache_ctx = xcache_get_ctx(cache);

	if (xcache_cleaner_prepare(cleaning_ctx) != 0) {
		return;
	}

	if (ocf_cleaner_run_prepare(cleaner, queue) != 0) {
		xcache_cleaner_complete(cleaning_ctx);
		return;
	}

	lock_idx = ocf_metadata_concurrency_next_idx(queue);
	ocf_metadata_start_shared_access(&cache->metadata.lock, lock_idx);

	cline_num = evicting_line_to_clean(xcache_ctx, cline, XCACHE_CLEAN_SIZE);

	ocf_metadata_end_shared_access(&cache->metadata.lock, lock_idx);

	if (cline_num == 0) {
		ocf_cleaner_run_complete(cleaner, CLEAN_INTERVAL);
		xcache_cleaner_complete(cleaning_ctx);
		return;
	}

	xcache_cleaner_set_end(cleaning_ctx, xcache_cleaner_run_complete, (void *)cleaner);
	xcache_cleaner_fill(cleaning_ctx, cline, cline_num);
	xcache_cleaner_clean(cache, queue, cleaning_ctx);
}
