/*
 * Copyright(c) 2012-2021 Intel Corporation
 * SPDX-License-Identifier: BSD-3-Clause-Clear
 */

#include "ocf_cache_priv.h"
#include "ocf_lru.h"
#include "metadata/metadata.h"
#include "utils/utils_alock.h"
#include "concurrency/ocf_cache_line_concurrency.h"

#include "xcache_cleaner.h"
#include "engine/xcache_engine_common.h"
#include "xcache_lru.h"
#include "evicting/evicting_ops.h"

static inline bool xcache_trylock_hash(struct xcache_io_context *ctx,
		ocf_core_id_t core_id, uint64_t core_line)
{
	if (ocf_metadata_hash_func(xcache_ctx_cache(ctx), core_line, core_id) ==
			ctx->hash)
		return true;

	return ocf_hb_cline_naked_trylock_wr(
			&xcache_ctx_cache(ctx)->metadata.lock,
			core_id, core_line);
}

static inline void xcache_unlock_hash(struct xcache_io_context *ctx,
		ocf_core_id_t core_id, uint64_t core_line)
{
	if (ocf_metadata_hash_func(xcache_ctx_cache(ctx), core_line, core_id) !=
			ctx->hash) {
		ocf_hb_cline_naked_unlock_wr(
				&xcache_ctx_cache(ctx)->metadata.lock,
				core_id, core_line);

	}
}

static inline bool xcache_eviction_lock(struct xcache_io_context *ctx,
		ocf_cache_line_t cache_line,
		ocf_core_id_t *core_id, uint64_t *core_line)

{
	struct ocf_alock *lock = ocf_cache_line_concurrency(xcache_ctx_cache(ctx));

	ocf_metadata_get_core_info(xcache_ctx_cache(ctx), cache_line,
		core_id, core_line);

	if (*core_id == ocf_core_get_id(xcache_ctx_core(ctx)) &&
			*core_line >= ctx->core_line_first &&
			*core_line <= ctx->core_line_last) {
		xcache_unlock_wr(lock, cache_line);
		return false;
	}

	if (!xcache_trylock_hash(ctx, *core_id, *core_line)) {
		xcache_unlock_wr(lock, cache_line);
		return false;
	}

	return true;
}

#define EVICT_RETRY_LOCK 3
static inline ocf_cache_line_t xcache_evict_line(struct xcache_io_context *io_ctx)
{
	ocf_cache_t cache = xcache_ctx_cache(io_ctx);
	xcache_context_t *ctx = xcache_get_ctx(cache);
	ocf_core_id_t core_id;
	uint64_t core_line;
	ocf_cache_line_t cline;
	int cline_num;
	int i;

	for (i = 0; i < EVICT_RETRY_LOCK; i++) {
		cline_num = evicting_line_to_evict(ctx, &cline, 1);
		if (cline_num == 0) {
			break;
		}
		if (xcache_eviction_lock(io_ctx,
					cline, &core_id, &core_line)) {
			goto found;
		}
	}
	return INVALID_LINE;

found:
	ocf_lru_invalidate(cache, cline, core_id, io_ctx->part_id);
	xcache_unlock_hash(io_ctx, core_id, core_line);

	xcache_map_cache_line(io_ctx, cline);
	return cline;
}

static inline ocf_cache_line_t xcache_free_list_line(ocf_cache_t cache, uint32_t list_idx,
		struct ocf_part *dst_part)
{
	struct ocf_part *free = &cache->free;
	struct ocf_alock *lock = ocf_cache_line_concurrency(cache);
	struct ocf_lru_list *list;
	ocf_cache_line_t cline;

	ocf_metadata_lru_wr_lock(&cache->metadata.lock, list_idx);

	list = ocf_lru_get_list(free, list_idx, true);
	for (cline = list->tail; cline != INVALID_LINE;
			cline = ocf_metadata_get_lru(cache, cline)->prev) {
		if (ocf_alock_trylock_entry_wr(lock, cline)) {
			break;
		}
	}

	if (cline != INVALID_LINE) {
		ocf_lru_repart_locked(cache, cline, free, dst_part);
	}

	ocf_metadata_lru_wr_unlock(&cache->metadata.lock,
			list_idx);

	return cline;
}

static ocf_cache_line_t xcache_get_free_line(struct xcache_io_context *ctx)
{
	struct ocf_part *dst_part;
	uint32_t start_idx, iter_idx;
	ocf_cache_line_t cline;

	dst_part = &xcache_ctx_cache(ctx)->user_parts[ctx->part_id].part;

	start_idx = xcache_ctx_queue(ctx)->lru_idx++ % OCF_NUM_LRU_LISTS;
	for (iter_idx = start_idx; iter_idx < OCF_NUM_LRU_LISTS; iter_idx++) {
		cline = xcache_free_list_line(xcache_ctx_cache(ctx), iter_idx, dst_part);
		if (cline != INVALID_LINE) {
			goto found;
		}
	}

	for (iter_idx = 0; iter_idx < start_idx; iter_idx++) {
		cline = xcache_free_list_line(xcache_ctx_cache(ctx), iter_idx, dst_part);
		if (cline != INVALID_LINE) {
			goto found;
		}
	}

	return INVALID_LINE;

found:
	ENV_BUG_ON(metadata_test_dirty(xcache_ctx_cache(ctx), cline));
	xcache_map_cache_line(ctx, cline);
	return cline;
}

ocf_cache_line_t xcache_get_cline(struct xcache_io_context *ctx)
{
	ocf_cache_t cache = xcache_ctx_cache(ctx);
	ocf_cache_line_t line;

	if (ocf_lru_num_free(cache) > 0) {
		line = xcache_get_free_line(ctx);
		if (line != INVALID_LINE) {
			return line;
		}
	}

	return xcache_evict_line(ctx);
}

void xcache_clean(struct xcache_io_context *io_ctx, uint32_t count)
{
	ocf_cache_t cache = xcache_ctx_cache(io_ctx);
	struct xcache_cleaning_ctx *cleaning_ctx = xcache_get_cleaning_ctx(cache);
	xcache_context_t *ctx = xcache_get_ctx(cache);
	ocf_cache_line_t cline[XCACHE_CLEAN_SIZE];
	int cline_num;

	unsigned i;
	unsigned lock_idx;

	if (ocf_mngt_cache_is_locked(cache))
		return;

	if (xcache_cleaner_prepare(cleaning_ctx) != 0) {
		return;
	}

	lock_idx = ocf_metadata_concurrency_next_idx(xcache_ctx_queue(io_ctx));
	ocf_metadata_start_shared_access(&cache->metadata.lock, lock_idx);

	count = count < XCACHE_CLEAN_SIZE ? count : XCACHE_CLEAN_SIZE;
	cline_num = evicting_line_to_clean(ctx, cline, count);
	xcache_cleaner_fill(cleaning_ctx, cline, cline_num);

	ocf_metadata_end_shared_access(&cache->metadata.lock, lock_idx);

	xcache_cleaner_clean(cache, xcache_ctx_queue(io_ctx), cleaning_ctx);
}
