#include "ocf/ocf.h"
#include "ocf_queue_priv.h"
#include "ocf_priv.h"
#include "ocf_cache_priv.h"

#include "ocf/xcache.h"
#include "xcache_queue.h"
#include "engine/xcache_engine.h"
#include "engine/xcache_engine_common.h"

#define ENTRY_MASK ((uintptr_t)0x7)

int xcache_queue_ctx_init(ocf_queue_t queue)
{
	struct xcache_queue_ctx *queue_ctx = env_malloc(sizeof(struct xcache_queue_ctx), 0);

	if (queue_ctx == NULL) {
		return -ENOMEM;
	}

	INIT_LIST_HEAD(&queue_ctx->backdev_io_list);
	queue_ctx->backdev_io_no = 0;

	INIT_LIST_HEAD(&queue_ctx->flush_io_list);
	queue_ctx->flush_io_no = 0;

	INIT_LIST_HEAD(&queue_ctx->xcache_io_list);
	queue_ctx->xcache_io_no = 0;

	queue_ctx->bf_data_num = 0;
	queue->priv1 = (void *)queue_ctx;
	return 0;
}

void xcache_queue_ctx_deinit(ocf_queue_t queue)
{
	env_free(queue->priv1);
	queue->priv1 = NULL;
}

static inline void set_entry_type(struct queue_entry *entry,
		enum entry_type type)
{
	if (type == OCF_REQ_ENTRY) {
		entry->list.next = (struct list_head *)(((uintptr_t)entry->list.next) | (uintptr_t)1);
		return;
	}

	entry->type = type;
}

static inline enum entry_type get_entry_type(struct queue_entry *entry)
{
	if (((uintptr_t)entry->list.next & ENTRY_MASK) != 0) {
		return OCF_REQ_ENTRY;
	}

	return entry->type;
}

static inline void list_add_entry(struct list_head *list, struct list_head *head)
{
	list_add(list, head);
}

static inline void list_add_tail_entry(struct list_head *list, struct list_head *head)
{
	struct list_head *prev = head->prev;
	uintptr_t bits = (uintptr_t)prev->next & ENTRY_MASK;

	list->next = head;
	list->prev = prev;

	prev->next = (struct list_head *)((uintptr_t)list | bits);
	head->prev = list;
}

static inline void list_pop_front_entry(struct queue_entry *entry)
{
	struct list_head *next = (struct list_head *)((uintptr_t)entry->list.next & ~ENTRY_MASK);

	entry->list.prev->next = next;
	next->prev = entry->list.prev;
}

static inline struct queue_entry *xcache_io_to_entry(struct xcache_io *io)
{
	return (struct queue_entry *)io;
}

static inline struct xcache_io *entry_to_xcache_io(struct queue_entry *entry)
{
	return (struct xcache_io *)entry;
}

static inline struct queue_entry *backdev_io_to_entry(struct xcache_backdev_io *base_io)
{
	return (struct queue_entry *)base_io;
}

static inline struct xcache_backdev_io *entry_to_backdev_io(struct queue_entry *entry)
{
	return (struct xcache_backdev_io *)entry;
}

static inline struct queue_entry *ocf_req_to_entry(struct ocf_request *req)
{
	return (struct queue_entry *)container_of(&req->list, struct queue_entry, list);
}

static inline struct ocf_request *entry_to_ocf_req(struct queue_entry *entry)
{
	return (struct ocf_request *)container_of(&(entry->list), struct ocf_request, list);
}

static void xcache_io_run(struct queue_entry *entry)
{
	struct xcache_io *io = entry_to_xcache_io(entry);

	io->io_if(io);

	xcache_io_put(io);
}

static void backdev_io_run(struct queue_entry *entry)
{
	struct xcache_backdev_io *base_io = entry_to_backdev_io(entry);
	struct xcache_io *io = base_io->xcache_io;

	base_io->io_res(base_io);

	xcache_io_put(io);
}

static void ocf_req_run(struct queue_entry *entry)
{
	struct ocf_request *req = entry_to_ocf_req(entry);

	if (req->ioi.io.handle)
		req->ioi.io.handle(&req->ioi.io, req);
	else
		ocf_io_handle(&req->ioi.io, req);
}

static struct queue_entry_ops {
	void (*entry_run)(struct queue_entry *entry);
} queue_entry_ops[] = {
	[XCACHE_IO_ENTRY] = {
		.entry_run = xcache_io_run,
	},
	[XCACHE_BACKDEV_IO_ENTRY] = {
		.entry_run = backdev_io_run,
	},
	[OCF_REQ_ENTRY] = {
		.entry_run = ocf_req_run,
	},
};

static inline void queue_entry_run(enum entry_type type, struct queue_entry *entry)
{
	queue_entry_ops[type].entry_run(entry);
}

static struct queue_entry *xcache_queue_pop_entry(ocf_queue_t q)
{
	unsigned long lock_flags = 0;
	struct queue_entry *entry;

	OCF_CHECK_NULL(q);

	env_spinlock_lock_irqsave(&q->io_list_lock, lock_flags);

	if (list_empty(&q->io_list)) {
		/* No items on the list */
		env_spinlock_unlock_irqrestore(&q->io_list_lock,
				lock_flags);
		return NULL;
	}

	/* Get the first request and remove it from the list */
	entry = list_first_entry(&q->io_list, struct queue_entry, list);

	env_atomic_dec(&q->io_no);
	list_pop_front_entry(entry);

	env_spinlock_unlock_irqrestore(&q->io_list_lock, lock_flags);

	OCF_CHECK_NULL(entry);

	return entry;
}

// overwrite ocf_queue_run_single
void ocf_queue_run_single(ocf_queue_t q)
{
	struct queue_entry *entry = NULL;

	OCF_CHECK_NULL(q);

	entry = xcache_queue_pop_entry(q);

	if (!entry)
		return;

	queue_entry_run(get_entry_type(entry), entry);
}

// only called by request in queue to avoid lock
struct xcache_backdev_io *xcache_queue_alloc_backdev_io(ocf_queue_t q)
{
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);
	struct xcache_backdev_io *io;

	if (queue_ctx->backdev_io_no == 0) {
		return env_malloc(sizeof(struct xcache_backdev_io), 0);
	}

	io = list_first_entry(&queue_ctx->backdev_io_list, struct xcache_backdev_io, free_list);
	list_del(&io->free_list);
	queue_ctx->backdev_io_no--;
	return io;
}

void xcache_queue_free_backdev_io(ocf_queue_t q, struct xcache_backdev_io *io)
{
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);

	if (queue_ctx->backdev_io_no > QUEUE_CACHE_SIZE) {
		env_free(io);
		return;
	}

	list_add_tail(&io->free_list, &queue_ctx->backdev_io_list);
	queue_ctx->backdev_io_no++;
}

struct xcache_io *xcache_queue_alloc_xcache_io(ocf_queue_t q)
{
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);
	struct xcache_io *io;

	if (queue_ctx->xcache_io_no == 0) {
		return env_malloc(sizeof(struct xcache_io), 0);
	}

	io = list_first_entry(&queue_ctx->xcache_io_list, struct xcache_io, queue_list);
	list_del(&io->queue_list);
	queue_ctx->xcache_io_no--;
	return io;
}

void xcache_queue_free_xcache_io(ocf_queue_t q, struct xcache_io *io)
{
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);

	if (queue_ctx->xcache_io_no > QUEUE_CACHE_SIZE) {
		env_free(io);
		return;
	}

	list_add_tail(&io->queue_list, &queue_ctx->xcache_io_list);
	queue_ctx->xcache_io_no++;
}

ctx_data_t *xcache_queue_alloc_line_data(ocf_queue_t q, ocf_cache_t cache)
{
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);

	if (queue_ctx->bf_data_num > 0) {
		return queue_ctx->bf_data[--queue_ctx->bf_data_num];
	} else {
		return ctx_data_alloc(cache->owner, (ocf_line_size(cache) + PAGE_SIZE - 1) / PAGE_SIZE);
	}
}

void xcache_queue_free_line_data(ocf_queue_t q, ocf_cache_t cache, ctx_data_t *data)
{
	struct xcache_queue_ctx *queue_ctx = xcache_get_queue_ctx(q);

	if (queue_ctx->bf_data_num < QUEUE_CACHE_SIZE) {
		queue_ctx->bf_data[queue_ctx->bf_data_num++] = data;
	} else {
		ctx_data_free(cache->owner, data);
	}
}

static void xcache_queue_push_entry(ocf_queue_t q, struct queue_entry *entry, bool at_head, bool allow_sync, enum entry_type type)
{
	ocf_cache_t cache = ocf_queue_get_cache(q);
	unsigned long lock_flags = 0;

	INIT_LIST_HEAD(&entry->list);

	env_atomic_set(&cache->last_access_ms,
			env_ticks_to_msecs(env_get_tick_count()));

	env_spinlock_lock_irqsave(&q->io_list_lock, lock_flags);

	if (at_head) {
		list_add_entry(&entry->list, &q->io_list);
	} else {
		list_add_tail_entry(&entry->list, &q->io_list);
	}
	// type set must under lock
	set_entry_type(entry, type);
	env_atomic_inc(&q->io_no);

	env_spinlock_unlock_irqrestore(&q->io_list_lock, lock_flags);

	ocf_queue_kick(q, allow_sync);
}

void xcache_queue_push_xcache_io_back(struct xcache_io *io, bool allow_sync)
{
	struct queue_entry *entry = xcache_io_to_entry(io);
	xcache_queue_push_entry(xcache_io_queue(io), entry, false, allow_sync, XCACHE_IO_ENTRY);
}

void xcache_queue_push_backdev_io_front(struct xcache_backdev_io *base_io, bool allow_sync)
{
	struct queue_entry *entry = backdev_io_to_entry(base_io);
	xcache_queue_push_entry(xcache_io_queue(base_io->xcache_io), entry, true, allow_sync, XCACHE_BACKDEV_IO_ENTRY);
}

// overwrite ocf_engine_push_req_front
void ocf_engine_push_req_front(struct ocf_request *req, bool allow_sync)
{
	struct queue_entry *entry = ocf_req_to_entry(req);
	xcache_queue_push_entry(req->io_queue, entry, true, allow_sync, OCF_REQ_ENTRY);
}

// overwrite ocf_engine_push_req_back
void ocf_engine_push_req_back(struct ocf_request *req, bool allow_sync)
{
	struct queue_entry *entry = ocf_req_to_entry(req);
	xcache_queue_push_entry(req->io_queue, entry, false, allow_sync, OCF_REQ_ENTRY);
}
